#include "raytracing.h"
#include "util.h"

void swap( double *a, double *b )
{
	double tmp;
	tmp = *a;
	*a  = *b;
	*b  = tmp;
}

bool block_intersect( struct Ray *ray, struct HitInfo *hitinfo, struct Object *obj )
{
	double Tnear, Tfar, T1, T2;
	struct Vec3 Min, Max;

	Min = obj->Min;
	Max = obj->Max;

	//set Tnear = - infinity, Tfar = infinity
	Tnear = -Infinity;
	Tfar  =  Infinity;

	//For each pair of planes P associated with X, Y, and Z do:
	
	//X planes
	if( ray->direction.x == 0 ) //if direction Xd = 0 then the ray is parallel to the X planes
	{
		//if origin Xo is not between the slabs ( Xo < Xl or Xo > Xh) then return false
		if( ray->origin.x < Min.x || ray->origin.x > Max.x )
			return 0;
	}
	else
	{
		T1 = ( Min.x - ray->origin.x ) / ray->direction.x;
		T2 = ( Max.x - ray->origin.x ) / ray->direction.x;
		
		if( T1 > T2 )      swap( &T1, &T2 );  //since T1 intersection with near plane
		if( T1 > Tnear)    Tnear =T1;       //want largest Tnear
		if( T2 < Tfar)     Tfar = T2;       //want smallest Tfar
		if( Tnear > Tfar ) return 0;
		if( Tfar < 0)      return 0;
	}	


	//Y planes
	if( ray->direction.y == 0 ) //if direction Yd = 0 then the ray is parallel to the Y planes
	{
		//if origin Yo is not between the slabs ( Yo < Yl or Yo > Yh) then return false
		if( ray->origin.y < Min.y || ray->origin.y > Max.y )
			return 0;
	}
	else
	{
		T1 = (Min.y - ray->origin.y ) / ray->direction.y;
		T2 = (Max.y - ray->origin.y ) / ray->direction.y;

		if( T1 > T2 )      swap( &T1, &T2 ); //since T1 intersection with near plane
		if( T1 > Tnear )   Tnear =T1;      //want largest Tnear
		if( T2 < Tfar )    Tfar = T2;        //want smallest Tfar
		if( Tnear > Tfar ) return 0;
		if( Tfar < 0 )     return 0;
	}	

	//Z planes
	if( ray->direction.z == 0 ) //if direction Zd = 0 then the ray is parallel to the Z planes
	{
		//if origin Xo is not between the slabs ( Zo < Zl or Zo > Zh) then return false
		if( ray->origin.z < Min.z || ray->origin.z > Max.z )
			return 0;
	}
	else
	{
		T1 = ( Min.z - ray->origin.z ) / ray->direction.z;
		T2 = ( Max.z - ray->origin.z ) / ray->direction.z;

		if( T1 > T2)       swap( &T1, &T2 ); //since T1 intersection with near plane
		if( T1 > Tnear )   Tnear =T1;        //want largest Tnear
		if( T2 < Tfar )    Tfar = T2;        //want smallest Tfar
		if( Tnear > Tfar ) return 0;
		if( Tfar < 0 )     return 0;
	}	

	//Tnear should be between 0 and saved distance
	if( Tnear < MachEps || Tnear > hitinfo->distance ) return 0;

	// We have an actual hit.  Fill in all the geometric information so
    // that the shader can shade this point.
	hitinfo->point  = vec3Add( ray->origin, vec3Mul( Tnear, Unit( ray->direction ) ) );
	hitinfo->distance = dist( ray->origin, hitinfo->point );
	
	if      ( fabs( hitinfo->point.x - Min.x ) < MachEps ) vec3Set( &hitinfo->normal, -1,  0,  0 ) ;
	else if ( fabs( hitinfo->point.y - Min.y ) < MachEps ) vec3Set( &hitinfo->normal,  0, -1,  0 );
	else if ( fabs( hitinfo->point.z - Min.z ) < MachEps ) vec3Set( &hitinfo->normal,  0,  0, -1 );
	else if ( fabs( hitinfo->point.x - Max.x ) < MachEps ) vec3Set( &hitinfo->normal,  1,  0,  0 );
	else if ( fabs( hitinfo->point.y - Max.y ) < MachEps ) vec3Set( &hitinfo->normal,  0,  1,  0 );
	else    vec3Set( &hitinfo->normal, 0, 0, 1 );

	hitinfo->object = obj;

	return 1;
}

struct Interval block_GetSlab( struct Object *obj, struct Vec3 *v )
{
    double ax = v->x * obj->Min.x;
	double ay = v->y * obj->Min.y;
    double az = v->z * obj->Min.z;
    double bx = v->x * obj->Max.x;
    double by = v->y * obj->Max.y;
    double bz = v->z * obj->Max.z;
    double a  = min( ax, bx ) + min( ay, by ) + min( az, bz );
    double b  = max( ax, bx ) + max( ay, by ) + max( az, bz ); 
	struct Interval I;
	//Interval( a, b ) / ( v * v );
	I.min = a; I.max = b;
	I = intervalDiv(I, vec3InnerProduct(*v, *v));
	return I;
}
