#include "raytracing.h"
#include "util.h"

bool triangle_intersect( struct Ray *ray, struct HitInfo *hitinfo, struct Object *obj )
{
	int axis, i, j;
	struct Vec3 N, A, B, C, P, Q;
	double W[3][3], M[3][3], AdjointW[3][3], detW, denom, s, d;

	A = obj->A;
	B = obj->B;
	C = obj->C;

    N = Unit( vec3CrossProduct( vec3Sub(A, B), vec3Sub(C, B) ) ); // Compute the normal vector.
    d = vec3InnerProduct( A, N );	// Distance from origin to plane of triangle.

    // Determine which axis is the "dominant" axis of the triangle.  That is, which
    // coordinate axis is closest to being orthogonal to the triangle.

    if( fabs(N.x) >= fabs(N.y) && fabs(N.x) >= fabs(N.z) ) axis = 0;
    if( fabs(N.y) >= fabs(N.x) && fabs(N.y) >= fabs(N.z) ) axis = 1;
    if( fabs(N.z) >= fabs(N.x) && fabs(N.z) >= fabs(N.y) ) axis = 2;
	if( axis>2 || axis < 0)
		return 0;

   // Set up the matrix for finding the barycentric coordinates.

	W[0][0] = A.x; W[0][1] = B.x; W[0][2] = C.x;
	W[1][0] = A.y; W[1][1] = B.y; W[1][2] = C.y;
	W[2][0] = A.z; W[2][1] = B.z; W[2][2] = C.z;
	W[axis][0] = 1.0;
	W[axis][1] = 1.0;
	W[axis][2] = 1.0;

    // Store the inverse of this matrix.  It will provide an efficient means of
    // finding the barycentric coordinates of the point of intersection with the
    // plane containing the triangle.

	detW = W[0][0] * ( W[1][1] * W[2][2] - W[1][2] * W[2][1] )
      - W[0][1] * ( W[1][0] * W[2][2] - W[1][2] * W[2][0] )
      + W[0][2] * ( W[1][0] * W[2][1] - W[1][1] * W[2][0] );

	AdjointW[0][0] = W[1][1] * W[2][2] - W[1][2] * W[2][1];
	AdjointW[0][1] = W[1][2] * W[2][0] - W[1][0] * W[2][2];
	AdjointW[0][2] = W[1][0] * W[2][1] - W[1][1] * W[2][0];

	AdjointW[1][0] = W[0][2] * W[2][1] - W[0][1] * W[2][2];
	AdjointW[1][1] = W[0][0] * W[2][2] - W[0][2] * W[2][0];
	AdjointW[1][2] = W[0][1] * W[2][0] - W[0][0] * W[2][1];

	AdjointW[2][0] = W[0][1] * W[1][2] - W[0][2] * W[1][1];
	AdjointW[2][1] = W[0][2] * W[1][0] - W[0][0] * W[1][2];
	AdjointW[2][2] = W[0][0] * W[1][1] - W[0][1] * W[1][0];

    for( i = 0; i < 3; i++ )
	    for( j = 0; j < 3; j++ ) 
			M[i][j] = AdjointW[j][i] / detW;

    // Compute the point of intersection with the plane containing the triangle.
    // Report a miss if the ray does not hit this plane.

    denom = vec3InnerProduct( ray->direction, N );
    if( fabs(denom) < 1.0E-4 ) return 0;
    s = ( d - vec3InnerProduct( ray->origin, N ) ) / denom;
    if( s <= 0.0 || s > hitinfo->distance ) return 0;
    P = vec3Add( ray->origin, vec3Mul( s, ray->direction ) );

    // Create a new vector that is a copy of P, but with one coordinate (corresponding
    // to the dominant exis) set to 1.  This is the right-hand-side of the equation for
    // the barycentric coordinates.

    Q = P;
    switch( axis )
        {
        case 0: Q.x = 1.0; break;
        case 1: Q.y = 1.0; break;
        case 2: Q.z = 1.0; break;
        }

    // Solve for the barycentric coordinates and check their sign.  The ray hits the triangle
    // if and only of all the barycentric coordinates are positive.

    if( M[0][0] * Q.x + M[0][1] * Q.y + M[0][2] * Q.z < 0.0 ) return 0;
    if( M[1][0] * Q.x + M[1][1] * Q.y + M[1][2] * Q.z < 0.0 ) return 0;
    if( M[2][0] * Q.x + M[2][1] * Q.y + M[2][2] * Q.z < 0.0 ) return 0;

    // We have an actual hit.  Fill in all the geometric information so
    // that the shader can shade this point.

    hitinfo->distance = s;
    hitinfo->point    = P; 
    hitinfo->normal   = N;
    hitinfo->object   = obj;
	return 1;
}

struct Interval triangle_GetSlab( struct Object *obj, struct Vec3 *v )
{
	const double a = vec3InnerProduct(*v, obj->A); //a = v * A;
    const double b = vec3InnerProduct(*v, obj->B); //b = v * B;
    const double c = vec3InnerProduct(*v, obj->C); //c = v * C;
	
	struct Interval I;
	I.min = OneMinusEps * min3( a, b, c );
	I.max = OnePlusEps  * max3( a, b, c );
	intervalDiv(I, vec3InnerProduct(*v, *v));
	return I;

    /*return Interval(
        OneMinusEps * min( a, b, c ),
        OnePlusEps  * max( a, b, c )
        ) / ( v * v );*/
}
